uses Graph, crt;

const
     cMaxCols = 16;
     cMaxColors = 32;
     cMaxRows = 10;

     cIntNone = 0;
     cIntBlack = 1;
     cIntWhite = 2;

{     cTones:array[1..10] of Integer = (262,294,330,349,392,440,493,523,587,659);}

type
     tCandidates = array[1..cMaxCols] of Longint;
     tGuess = array[1..cMaxCols] of Byte;

     tTurn = record
       guess : tGuess;
       candid : tCandidates;
       clrdata : array[1..cMaxColors] of record
                                              maxcnt,mincnt:shortint;
                                            end;
       blacks : Byte;
       whites : Byte;
       interp : Longint;
       assumeand, assumeset : Longint;
     end;

     tGame = array[0..cMaxRows] of tTurn;

var game:tGame;
    lastinterp:longint;
    maxcols:byte;
    maxcolors:byte;

{ Spocita jednickove bity v longintu }
function sumBites(l:longint):byte;
begin
  l := (l and $55555555)+(l shr 1 and $55555555);
  l := (l and $33333333)+(l shr 2 and $33333333);
  l := (l and $0F0F0F0F)+(l shr 4 and $0F0F0F0F);
  l := (l and $00FF00FF)+(l shr 8 and $00FF00FF);
  l := (l and $0000FFFF)+(l shr 16 and $0000FFFF);
  sumBites := byte(l);
end;

{ Vrati TRUE, je-li color v turn-tahu kandidatem na positiontou pozici.
  T.zn. je-li DUSLEDKEM tohoto tahu jeho kandidatstvi }
function isCandidate(color, position, turn:byte):boolean;
begin
  isCandidate := (game[turn].candid[position] shr (color-1)) and 1=1;
end;

{ Vrati bity interpretace prislusne pozice v turn-tem tahu }
function maskInterp(position, turn:byte):byte;
begin
  maskInterp := (game[turn].interp shr ((position-1) shl 1)) and 3;
end;

{ Najde nasledujici ohodnoceni, ktere odpovida poctu cernych a bilych
  ohodnocujicich cudliku a zaroven splnuje predpoklady z predchozich
  tahu }
function findNextInterpretation(turn:byte):boolean;
var blk,wht:byte;
begin
  findNextInterpretation := false;
  with game[turn] do begin
    repeat
      inc(interp);
      { upravime na trojkove scitani }
      while ((interp and (interp shl 1)) shr 1) and $55555555<>0 do
        interp := interp + (((interp and (interp shl 1)) shr 1) and $55555555);
      blk := sumBites((interp and (interp shr 1 xor $55555555)) and $55555555);
      wht := sumBites(((interp xor $55555555) and (interp shr 1)) and $55555555);
      if interp>lastinterp then
        exit;
    until (blk=blacks) and (wht=whites) and (interp and assumeand=assumeset);
  end;
  findNextInterpretation := true;
end;

function min(i,j:integer):integer;
begin
  if i<j then
    min := i
  else
    min := j;
end;

function max(i,j:integer):integer;
begin
  if i>j then
    max := i
  else
    max := j;
end;

{ Na zaklade ohodnoceni spocita dusledky tohoto tahu, tudiz prepocita
  candid a clrdata }
procedure processInterpretation(turn:byte);
var i,j:integer;
    l:longint;
begin
  with game[turn] do begin
    candid := game[turn-1].candid;
    for i:=1 to maxcolors do begin
      clrdata[i].maxcnt := maxcols;
      clrdata[i].mincnt := 0;
    end;
    l := interp;
    { nejprve projedeme bile }
    for i:=1 to maxcols do begin
      if l and 3 = 2 then begin
        inc(clrdata[guess[i]].mincnt);
        candid[i] := candid[i] and (1 shl (guess[i] - 1));
      end;
      l := l shr 2;
    end;
    l := interp;
    { pak cerne }
    for i:=1 to maxcols do begin
      if l and 3 = 1 then begin
        inc(clrdata[guess[i]].mincnt);
        candid[i] := candid[i] and not(1 shl (guess[i] -1));
      end;
      l := l shr 2;
    end;
    l := interp;
    { pak zadne }
    for i:=1 to maxcols do begin
      if l and 3 = 0 then begin
        clrdata[guess[i]].maxcnt := clrdata[guess[i]].mincnt;
        candid[i] := candid[i] and not(1 shl (guess[i] -1));
      end;
      l := l shr 2;
    end;
    for i:=1 to maxcolors do begin
      clrdata[i].maxcnt := min(clrdata[i].maxcnt, game[turn-1].clrdata[i].maxcnt);
      clrdata[i].mincnt := max(clrdata[i].mincnt, game[turn-1].clrdata[i].mincnt);
      if clrdata[i].maxcnt=0 then
        for j:=1 to maxcols do
          candid[j] := candid[j] and not(1 shl (i-1));
    end;
  end;
end;

{ Vypocte assumpce ohodnoceni turn-teho tahu na zaklade dat z tahu predcho-
  ziho }
procedure setAssumptions(turn:byte);
var i:integer;
begin
  with game[turn] do begin
    assumeand := 0;
    assumeset := 0;
    for i:=maxcols downto 1 do begin
      assumeand := assumeand shl 2;
      assumeset := assumeset shl 2;
      if game[turn-1].candid[i] = 1 shl (guess[i]-1) then begin
        assumeand := assumeand or 3;
        assumeset := assumeset or 2;
      end else
        if not isCandidate(guess[i], i, turn-1) then
          if game[turn-1].clrdata[guess[i]].maxcnt=0 then
            assumeand := assumeand or 3
          else
            assumeand := assumeand or 2;
    end;
  end;
end;

{ Otestuje, zda je na kazdou pozici alespon jeden kandidat }
function  validAfterProcess(turn:byte):boolean;
var i:integer;
begin
  validAfterProcess := false;
  with game[turn] do begin
    for i:=1 to maxcols do begin
      if candid[i]=0 then
        exit;
    end;
  end;
  validAfterProcess := true;
end;

{ Pokusi se najit nasledujici logicky pasujici interpretaci, pokud
  nenalezne, pokusi se rekurzivne vnorit do predchoziho tahu }
function shiftInterpretation(turn:byte):boolean;
begin
  shiftInterpretation := false;
  if turn=0 then begin
    if game[0].interp=$FFFFFFFF then begin
      game[0].interp:=0;
      shiftInterpretation := true;
    end;
    exit;
  end;
  repeat
    if not(findNextInterpretation(turn)) then begin
      if not(shiftInterpretation(turn-1)) then
        exit;
      game[turn].interp := $FFFFFFFF;
      findNextInterpretation(turn);
    end;
    processInterpretation(turn);
    setAssumptions(turn+1);
  until validAfterProcess(turn);
  shiftInterpretation := true;
end;

{ Zkusi dosadit cudlik na pos-tou pozici turn-tem tahu, pricemz zacne
  od beg-te barvy (kvuli rotaci) }
function guessPosition(turn, pos, beg:byte):boolean;
var i,j:integer;
begin
  if (pos=0) then begin
    guessPosition := true;
    for i:=1 to maxcolors do
      if game[turn-1].clrdata[i].mincnt>0 then
        guessPosition := false;
    exit;
  end;
  for j:=1 to maxcolors do begin
    i := (j+beg-2) mod maxcolors + 1;
    if (isCandidate(i, pos, turn-1)) and (game[turn-1].clrdata[i].maxcnt>0) then begin
      dec(game[turn-1].clrdata[i].maxcnt);
      dec(game[turn-1].clrdata[i].mincnt);
      if guessPosition(turn, pos-1, i+1) then begin
        inc(game[turn-1].clrdata[i].maxcnt);
        inc(game[turn-1].clrdata[i].mincnt);
        game[turn].guess[pos] := i;
        guessPosition := true;
        exit;
      end;
      inc(game[turn-1].clrdata[i].maxcnt);
      inc(game[turn-1].clrdata[i].mincnt);
    end;
  end;
  guessPosition := false;
end;

{ Pokusi se vymyslet turn-ty tah }
function guessTurn(turn:byte):boolean;
begin
  guessTurn := guessPosition(turn, maxcols, 1);
end;

{ Pokusi se zahrat turn-ty tah }
function play(turn:byte):boolean;
begin
  play := true;
  while shiftInterpretation(turn-1) do begin
    if guessTurn(turn) then
      exit;
  end;
  play := false;
end;

var toguess:tGuess;

procedure drawGadget(X,Y:Word; Color:Byte; Lesk:Boolean);
begin
  SetFillStyle(1, Color);
  SetColor(Black);
  SetLineStyle(SolidLn,0,3);
  Circle(X+1, Y+1, 10);
  SetLineStyle(SolidLn,0,1);
  SetColor(Color);
  PieSlice(X, Y, 0, 360, 10);
  SetColor(Black);
  Circle(X,Y,10);
  if Lesk then begin
    SetColor(White);
    Line(X-3, Y-2, X-1, Y-2);
    Line(X-2, Y-3, X-2, Y-1);
  end;
end;

procedure drawStick(X,Y:Word; Color:Byte);
begin
  SetLineStyle(SolidLn, 0, 3);
  SetColor(Color);
  SetFillStyle(SolidFill, Color);
  Line(X, Y, X+5, Y-10);
  PieSlice(X+5, Y-10, 0, 360, 3);
  SetLineStyle(SolidLn, 0, 1);
end;

procedure beep(t:integer);
begin
  sound(cTones[t]);
  delay(50);
  nosound;
end;

procedure showTurn(turn:byte);
var i:integer;
begin
  for i:=1 to maxcols do
    drawGadget(20+i*30, 37+turn*28, game[turn].guess[i], True);
end;

procedure showEvaluation(turn:byte);
var i,j:integer;
begin
  i:=1;
  for j:=1 to game[turn].whites do begin
    drawStick(230+i*20, 37+turn*28, White);
    inc(i);
  end;
  for j:=1 to game[turn].blacks do begin
    drawStick(230+i*20, 37+turn*28, LightGray);
    inc(i);
  end;
end;

{ ohodnoti turn-ty tah }
procedure evaluateTurn(turn:byte);
var i,j:integer;
    tg:tGuess;
    g:tGuess;
begin
  tg := toguess;
  g := game[turn].guess;
  game[turn].blacks := 0;
  game[turn].whites := 0;
  for i:=1 to maxcols do
    if tg[i]=game[turn].guess[i] then begin
      inc(game[turn].whites);
      tg[i] := 0;
      g[i] := 0;
    end;
  for i:=1 to maxcols do
    if g[i]<>0 then
      for j:=1 to maxcols do
        if tg[j]=g[i] then begin
          tg[j] := 0;
          g[i] := 0;
          inc(game[turn].blacks);
          break;
        end;
end;

{ Hlavni smycka hry... Vyplni imaginarni 0.ty tah a hraje dokud se prislusna
  kombinace nenajde }
function gameLoop:word;
var turn:word;
    g:word;
begin
  lastinterp := 0;
  for g:=1 to maxcols do
    lastinterp := (lastinterp shl 2) or 2;
  game[0].interp := $FFFFFFFF;
  for g:=1 to maxcols do begin
    game[0].candid[g] := (1 shl maxcolors)-1;
    game[0].assumeand := 0;
    game[0].assumeset := 0;
  end;
  for g:=1 to maxcolors do begin
    game[0].clrdata[g].maxcnt := maxcols;
    game[0].clrdata[g].mincnt := 0;
  end;
  turn := 0;
  gameloop := 0;
  repeat
    inc(turn);
    if not(play(turn)) then begin
      writeln('cheater!');
      exit;
    end;
    showTurn(turn);
    evaluateTurn(turn);
    showEvaluation(turn);
    beep(turn);
    game[turn].interp := $FFFFFFFF;
    game[turn].assumeand := 0;
    game[turn].assumeset := 0;
  until game[turn].whites=maxcols;
  for g:=turn-1 downto 1 do begin
    delay(30);
    beep(g);
  end;
  gameloop := turn;
end;

procedure redrawScreen;
var i,j:integer;
    s:string[3];
begin
  clearDevice;
  SetFillStyle(SolidFill, DarkGray);
  Bar(0,0,GetMaxX,GetMaxY);
  SetColor(White);
  Rectangle(0,0,GetMaxX,GetMaxY);
  SetTextStyle(SmallFont, HorizDir, 6);
  SetTextJustify(LeftText, TopText);
  OutTextXY(10, 10, 'Povolene barvy:');
  SetTextJustify(CenterText, CenterText);
  SetTextStyle(SmallFont, HorizDir, 4);
  for i:=1 to maxcolors do begin
    drawGadget(GetMaxX-i*30, 20, maxcolors-i+1, False);
    SetColor(White);
    OutTextXY(GetMaxX-i*30, 20, Char(maxcolors-i+65));
  end;
  SetColor(White);
  Line(0,40,GetMaxX,40);
  SetFillStyle(SolidFill, 0);
  Bar(423, 53, GetMaxX-7, 253);
  SetFillStyle(SolidFill, LightGray);
  Bar(420, 50, GetMaxX-10, 250);
  SetColor(Black);
  Rectangle(420,50,GetMaxX-10,250);
  Line(420,72,GetMaxX-10,72);
  SetTextJustify(CenterText, TopText);
  SetTextStyle(SmallFont, HorizDir, 5);
  SetColor(Black);
  OutTextXY(525, 53, 'Informace + Napoveda');

  SetTextJustify(LeftText, TopText);
  OutTextXY(430, 75, '[N]   - nova hra');
  OutTextXY(430, 90, '[P]   - zmena parametru');
  OutTextXY(430, 105, '[ESC] - konec');

  SetTextJustify(CenterText, CenterText);
  SetTextStyle(0, HorizDir, 0);
  for i:=1 to cMaxRows do begin
    SetColor(White);
    Str(i, s);
    OutTextXY(20,39+i*28,s);
    for j:=1 to maxcols do begin
      drawGadget(20+j*30, 37+i*28, 0, False);
      SetColor(Black);
      SetFillStyle(SolidFill, 0);
      PieSlice(230+j*20, 37+i*28, 0, 360, 4);
    end;
  end;
end;

procedure newGame;
var i:integer;
    ch:char;
begin
  SetTextStyle(SmallFont, HorizDir, 5);
  SetTextJustify(LeftText, TopText);
  SetColor(White);
  OutTextXY(430, 260, 'Hadana kombinace: ');
  for i:=1 to maxcols do
    drawGadget(GetMaxX-i*30, 300, 0, False);
  SetTextStyle(0, HorizDir, 0);
  SetTextJustify(CenterText, CenterText);
  SetColor(White);
  for i:=1 to maxcols do begin
    SetColor(White);
    OutTextXY(GetMaxX-(maxcols-i+1)*30, 300, '?');
    repeat
      ch := upcase(readkey);
    until (ch=#27) or ((ch>'@') and (ord(ch)<65+maxcolors));
    if ch=#27 then begin
      redrawScreen;
      exit;
    end;
    toguess[i] := ord(ch)-64;
    beep(toguess[i]);
    drawGadget(GetMaxX-(maxcols-i+1)*30, 300, toguess[i], True);
  end;

  i := gameLoop;

  SetTextStyle(0, HorizDir, 1);
  SetTextJustify(CenterText, CenterText);
  SetColor(White);
  OutTextXY(125, 37+(i+1)*28, 'Stisknete klavesu...');
  while keypressed do readkey;
  if readkey=#0 then readkey;
  redrawScreen;
end;

procedure changeParams;
var ch:char;
begin
  SetTextJustify(LeftText, TopText);
  SetTextStyle(SmallFont, HorizDir, 5);
  SetColor(Black);
  OutTextXY(430, 150, 'Pocet mist [1-6]:');
  repeat
    ch := readkey;
  until (ch>='1') and (ch<='6');
  maxcols := ord(ch)-48;
  OutTextXY(430, 165, 'Pocet barev [3-9]:');
  repeat
    ch := readkey;
  until (ch>='3') and (ch<='9');
  maxcolors := ord(ch)-48;
  redrawScreen;
end;

procedure mainLoop;
var ch:char;
begin
  repeat
    repeat
      ch := upcase(readkey);
    until (ch=#27) or (ch='P') or (ch='N');
    case ch of
      'P':changeParams;
      'N':newGame;
    end;
  until ch=#27;
end;

var gd, gm:integer;

begin
  gd := Ega; gm := EgaHi;
  initGraph(gd, gm, 'c:\bp\bgi');

  maxcols := 5;
  maxcolors := 8;

  redrawScreen;

  mainLoop;

  closeGraph;
end.
