/*
 *	SynCom -- Syntax Trees
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "syncom.h"

struct node *
new_node(int id, ...)
{
  struct node *n = xmalloc(sizeof(struct node));
  va_list args;
  struct node **last, *m;

  va_start(args, id);
  n->children = n->sibling = NULL;
  n->leaf = 0;
  n->u.id = id;
  last = &n->children;
  while (m = va_arg(args, struct node *))
    {
      *last = m;
      last = &m->sibling;
    }
  va_end(args);
  return n;
}

struct node *
new_leaf(struct keyword *w)
{
  struct node *n = xmalloc(sizeof(struct node));

  n->children = n->sibling = NULL;
  n->leaf = 1;
  n->u.w = w;
  return n;
}

static void
do_dump_tree(struct node *n, int k)
{
  int i;

  for(i=0; i<k; i++)
    putchar(' ');
  if (!n->leaf)
    {
      printf("[rule %d]\n", n->u.id);
      for(n=n->children; n; n=n->sibling)
	do_dump_tree(n, k+2);
    }
  else
    printf("<%s>\n", n->u.w->w);
}

void
dump_tree(struct node *n)
{
  if (n)
    do_dump_tree(n, 0);
  else
    puts("<NULL>");
}
